@@ -14,12 +14,16 @@ from requests import post, put
 from requests.exceptions import HTTPError
 from yellowdog_client.model import (
     AccountAllowance,
+    AddApplicationResponse,
     AddConfiguredWorkerPoolResponse,
     AllowanceLimitEnforcement,
     AllowanceResetType,
+    ApiKey,
+    Application,
     AwsFleetComputeSource,
     AwsFleetPurchaseOption,
     CloudProvider,
+    Group,
     ImageOsType,
     Keyring,
     MachineImage,
@@ -30,6 +34,7 @@ from yellowdog_client.model import (
     RequirementsAllowance,
     SourceAllowance,
     SourcesAllowance,
+    User,
 )
 from yellowdog_client.model.exceptions import InvalidRequestException
 
@@ -39,6 +44,14 @@ from yellowdog_cli.utils.entity_utils import (
     find_compute_requirement_template_id_by_name,
     find_compute_source_template_id_by_name,
     find_image_family_reference_by_name,
+    get_application_groups,
+    get_application_id_by_name,
+    get_group_id_by_name,
+    get_group_name_by_id,
+    get_role_id_by_name,
+    get_role_name_by_id,
+    get_user_by_name_or_id,
+    get_user_groups,
     remove_allowances_matching_description,
 )
 from yellowdog_cli.utils.interactive import confirmed
@@ -59,6 +72,8 @@ from yellowdog_cli.utils.settings import (
     PROP_DESCRIPTION,
     PROP_EFFECTIVE_FROM,
     PROP_EFFECTIVE_UNTIL,
+    PROP_GROUPS,
+    PROP_ID,
     PROP_IMAGE,
     PROP_IMAGE_ID,
     PROP_IMAGES_ID,
@@ -70,15 +85,20 @@ from yellowdog_cli.utils.settings import (
     PROP_RANGE,
     PROP_REQUIREMENT_CREATED_FROM,
     PROP_RESOURCE,
+    PROP_ROLES,
     PROP_SOURCE,
     PROP_SOURCE_CREATED_FROM,
     PROP_SOURCES,
     PROP_TITLE,
     PROP_TYPE,
     PROP_UNITS,
+    RN_ADD_APPLICATION_REQUEST,
+    RN_ADD_GROUP_REQUEST,
     RN_ALLOWANCE,
+    RN_APPLICATION,
     RN_CONFIGURED_POOL,
     RN_CREDENTIAL,
+    RN_GROUP,
     RN_IMAGE_FAMILY,
     RN_KEYRING,
     RN_NAMESPACE_POLICY,
@@ -87,6 +107,9 @@ from yellowdog_cli.utils.settings import (
     RN_SOURCE_TEMPLATE,
     RN_STORAGE_CONFIGURATION,
     RN_STRING_ATTRIBUTE_DEFINITION,
+    RN_UPDATE_APPLICATION_REQUEST,
+    RN_UPDATE_GROUP_REQUEST,
+    RN_USER,
 )
 from yellowdog_cli.utils.wrapper import ARGS_PARSER, CLIENT, CONFIG_COMMON, main_wrapper
 from yellowdog_cli.utils.ydid_utils import YDIDType, get_ydid_type
@@ -160,6 +183,12 @@ def create_resources(
                 create_attribute_definition(resource, resource_type)
             elif resource_type == RN_NAMESPACE_POLICY:
                 create_namespace_policy(resource)
+            elif resource_type == RN_GROUP:
+                create_group(resource)
+            elif resource_type == RN_APPLICATION:
+                create_application(resource)
+            elif resource_type == RN_USER:
+                update_user(resource)
             else:
                 print_error(f"Unknown resource type '{resource_type}'")
         except Exception as e:
@@ -957,6 +986,248 @@ def create_namespace_policy(resource: Dict):
     )
 
 
+def create_group(resource: Dict):
+    """
+    Create or update a group. Will also add or remove roles specified
+    by their names or IDs.
+    """
+    try:
+        name = resource[PROP_NAME]
+    except KeyError as e:
+        raise Exception(f"Expected property to be defined ({e})")
+
+    roles: Optional[List[str]] = resource.pop(PROP_ROLES, None)
+
+    if roles is not None:
+        # Convert role names to IDs
+        new_role_ids = set()
+        for role_name in roles:
+            role_id = get_role_id_by_name(CLIENT, role_name)
+            if role_id is None:
+                print_warning(f"Role name '{role_name}' not found ... ignoring")
+            else:
+                new_role_ids.add(role_id)
+
+    def update_roles(group: Group):
+        """
+        Helper function to add/remove roles from a group.
+        """
+        if roles is None:
+            return
+
+        current_role_ids = {role.role.id for role in group.roles}
+
+        role_ids_to_remove = current_role_ids - new_role_ids
+        for role_id in role_ids_to_remove:
+            CLIENT.account_client.remove_role_from_group(group.id, role_id)
+            print_log(
+                f"Removed Role '{get_role_name_by_id(CLIENT, role_id)}' "
+                f"from Group ({role_id})"
+            )
+
+        role_ids_to_add = new_role_ids - current_role_ids
+        for role_id in role_ids_to_add:
+            CLIENT.account_client.add_role_to_group(group.id, role_id)
+            print_log(
+                f"Added Role '{get_role_name_by_id(CLIENT, role_id)}' "
+                f"to Group ({role_id})"
+            )
+
+    def add_group():
+        """
+        Helper function to add a new group and its roles.
+        """
+        group: Group = CLIENT.account_client.add_group(
+            _get_model_object(RN_ADD_GROUP_REQUEST, resource)
+        )
+        print_log(f"Created Group '{group.name}' ({group.id})")
+        update_roles(group)
+
+    def update_group(group_id: str):
+        """
+        Helper function to update an existing group, including updating
+        its roles.
+        """
+        group: Group = CLIENT.account_client.update_group(
+            group_id, _get_model_object(RN_UPDATE_GROUP_REQUEST, resource)
+        )
+        print_log(f"Updated Group '{group.name}' ({group.id})")
+        update_roles(group)
+
+    # Main logic
+    group_id = get_group_id_by_name(CLIENT, name)
+    if group_id is None:
+        add_group()
+    else:
+        update_group(group_id)
+
+
+def create_application(resource: Dict):
+    """
+    Create or update an application. Will also add or remove groups specified
+    by their names or IDs.
+    """
+    try:
+        name = resource[PROP_NAME]
+    except KeyError as e:
+        raise Exception(f"Expected property to be defined ({e})")
+
+    groups: Optional[List[str]] = resource.pop(PROP_GROUPS, None)
+
+    if groups is not None:
+        # Convert group names to IDs
+        new_group_ids = set()
+        for group_name in groups:
+            app_id = get_group_id_by_name(CLIENT, group_name)
+            if app_id is None:
+                print_warning(f"Group name '{group_name}' not found ... ignoring")
+            else:
+                new_group_ids.add(app_id)
+
+    def update_groups(app: Application):
+        """
+        Helper function to add/remove groups from an application.
+        """
+        if groups is None:
+            return
+
+        current_group_ids = {
+            group.id for group in get_application_groups(CLIENT, app.id)
+        }
+
+        group_ids_to_remove = current_group_ids - new_group_ids
+        for group_id in group_ids_to_remove:
+            CLIENT.account_client.remove_application_from_group(group_id, app.id)
+            print_log(
+                f"Removed Group '{get_group_name_by_id(CLIENT, group_id)}' "
+                f"from Application ({group_id})"
+            )
+
+        group_ids_to_add = new_group_ids - current_group_ids
+        for group_id in group_ids_to_add:
+            CLIENT.account_client.add_application_to_group(group_id, app.id)
+            print_log(
+                f"Added Group '{get_group_name_by_id(CLIENT, group_id)}' "
+                f"to Application ({group_id})"
+            )
+
+    def show_key_and_secret(api_key: ApiKey):
+        """
+        Helper function to display the app key and secret.
+        """
+        print_log(f"Application Key ID     = '{api_key.id}'", override_quiet=True)
+        print_log(f"Application Key Secret = '{api_key.secret}'", override_quiet=True)
+
+    def add_application():
+        """
+        Helper function to add a new application and its groups.
+        """
+        app_response: AddApplicationResponse = CLIENT.account_client.add_application(
+            _get_model_object(RN_ADD_APPLICATION_REQUEST, resource)
+        )
+        app = app_response.application
+        print_log(f"Created Application '{app.name}' ({app.id})")
+        show_key_and_secret(app_response.apiKey)
+        update_groups(app)
+
+    def update_application(app_id: str):
+        """
+        Helper function to update an existing application, including updating
+        its groups.
+        """
+        app: Application = CLIENT.account_client.update_application(
+            app_id, _get_model_object(RN_UPDATE_APPLICATION_REQUEST, resource)
+        )
+        print_log(f"Updated Application '{app.name}' ({app.id})")
+        update_groups(app)
+
+        if ARGS_PARSER.regenerate_app_keys:
+            print_log("Regenerating Application key and secret")
+            api_key = CLIENT.account_client.regenerate_application_api_key(app_id)
+            if api_key is None:
+                print_error("New API key/secret not returned")
+            else:
+                show_key_and_secret(api_key)
+
+    # Main logic
+    app_id = get_application_id_by_name(CLIENT, name)
+    if app_id is None:
+        add_application()
+    else:
+        update_application(app_id)
+
+
+def update_user(resource: Dict):
+    """
+    Update a user. Will also add or remove groups specified
+    by their names or IDs.
+    """
+
+    name = resource.get(PROP_NAME)
+    id = resource.get(PROP_ID)
+
+    if name is None and id is None:
+        raise Exception(
+            f"Expected one of '{PROP_NAME}', '{PROP_ID}' to be defined for "
+            f"resource '{RN_USER}'"
+        )
+
+    groups: Optional[List[str]] = resource.pop(PROP_GROUPS, None)
+
+    if groups is not None:
+        # Convert group names to IDs
+        new_group_ids = set()
+        for group_name in groups:
+            app_id = get_group_id_by_name(CLIENT, group_name)
+            if app_id is None:
+                print_warning(f"Group name '{group_name}' not found ... ignoring")
+            else:
+                new_group_ids.add(app_id)
+
+    def update_groups(user: User):
+        """
+        Helper function to add/remove groups from a user.
+        """
+        if groups is None:
+            return
+
+        current_group_ids = {group.id for group in get_user_groups(CLIENT, user.id)}
+
+        group_ids_to_remove = current_group_ids - new_group_ids
+        for group_id in group_ids_to_remove:
+            CLIENT.account_client.remove_user_from_group(group_id, user.id)
+            print_log(
+                f"Removed Group '{get_group_name_by_id(CLIENT, group_id)}' "
+                f"from User ({group_id})"
+            )
+
+        group_ids_to_add = new_group_ids - current_group_ids
+        for group_id in group_ids_to_add:
+            CLIENT.account_client.add_user_to_group(group_id, user.id)
+            print_log(
+                f"Added Group '{get_group_name_by_id(CLIENT, group_id)}' "
+                f"to User ({group_id})"
+            )
+
+    # Main logic: try name then ID if present
+    user = None
+    if name is not None:
+        user = get_user_by_name_or_id(CLIENT, name)
+    if user is None and id is not None:
+        user = get_user_by_name_or_id(CLIENT, id)
+
+    if user is None:
+        print_warning(
+            f"User '{name}' not found; Users cannot be created using "
+            "the CLI, please use the YellowDog Portal"
+        )
+    elif groups is not None:
+        print_log(f"Updating Groups for User '{name}' ({user.id})")
+        update_groups(user)
+    else:
+        print_log(f"Nothing to do for User '{name}' ({user.id})")
+
+
 # Entry point
 if __name__ == "__main__":
     main()
